from typing import Any

from flamapy.metamodels.fm_metamodel.models.feature_model import (
    FeatureModel,
    Constraint,
    Feature,
    Relation,
)

from flamapy.metamodels.pysat_metamodel.transformations.fm_to_pysat import FmToPysat
from ..models.pysat_diagnosis_model import DiagnosisModel


class FmToDiagPysat(FmToPysat):
    @staticmethod
    def get_source_extension() -> str:
        return 'fm'

    @staticmethod
    def get_destination_extension() -> str:
        return 'pysat_diagnosis'

    def __init__(self, source_model: FeatureModel) -> None:
        super().__init__(source_model)
        self.destination_model = DiagnosisModel()

    def add_root(self, feature: Feature) -> None:
        var = self.destination_model.variables.get(feature.name)
        if var is None:
            raise KeyError(f'Feature {feature.name} not found in the model')

        self.destination_model.add_clause([var])
        self.destination_model.add_clause_to_map(str(feature), [[var]])

    #def _store_constraint_relation(self, relation: Relation, clauses: List[List[int]]) -> None:
    #    for clause in clauses:
    #        self.destination_model.add_clause(clause)
    #    self.destination_model.add_clause_to_map(str(relation), clauses)
    def add_relation(self, relation: Relation) -> None:
        if relation.is_mandatory():
            clauses = self._add_mandatory_relation(relation)
        elif relation.is_optional():
            clauses = self._add_optional_relation(relation)
        elif relation.is_or():
            clauses = self._add_or_relation(relation)
        elif relation.is_alternative():
            clauses = self._add_alternative_relation(relation)
        else:
            clauses = self._add_constraint_relation(relation)
        self._store_constraint_clauses(clauses)
        self.destination_model.add_clause_to_map(str(relation), clauses)

    def add_constraint(self, ctc: Constraint) -> None:
        def get_term_variable(term: Any) -> int:
            negated = False
            if term.startswith('-'):
                term = term[1:]
                negated = True

            var = self.destination_model.get_variable(term)

            if negated:
                return -var
            return var

        ctc_clauses = []
        clauses = ctc.ast.get_clauses()
        for clause in clauses:
            clause_variables = list(map(get_term_variable, clause))
            ctc_clauses.append(clause_variables)
            self.destination_model.add_clause(clause_variables)

        self.destination_model.add_clause_to_map(str(ctc), ctc_clauses)


from typing import Any, cast, Callable, Optional

from flamapy.core.models import VariabilityModel
from flamapy.core.exceptions import FlamaException
from flamapy.core.operations.metrics_operation import Metrics
from flamapy.metamodels.pysat_metamodel.models import PySATModel
from flamapy.metamodels.pysat_metamodel import operations as sat_operations


def metric_method(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to mark a method as a metric method.
    It has the value of the measure, it can also have a size and a ratio.
    Example:
        property name: Abstract Features.
        description: The description of the property.
        value (optional): the list of abstract features.
        size (optional): the length of the list.
        ratio (optional): the percentage of abstract features with regards the total number
                            of features.
    """
    if not hasattr(func, '_is_metric_method'):
        setattr(func, '_is_metric_method', True)
    return func


class PySATMetrics(Metrics):

    def __init__(self) -> None:
        super().__init__()
        self.model: Optional[VariabilityModel] = None
        self.result: list[dict[str, Any]] = []
        self._model_type_extension = "pysat"
        self._features: dict[int, str] = {}
        self._common_features: list[Any] = []
        self._dead_features: list[Any] = []

    @property
    def model_type_extension(self) -> str:
        return self._model_type_extension

    @model_type_extension.setter
    def model_type_extension(self, ext: str) -> None:
        self._model_type_extension = ext

    def get_result(self) -> list[dict[str, Any]]:
        return self.result

    def calculate_metamodel_metrics(self, model: VariabilityModel) -> list[dict[str, Any]]:
        self.model = cast(PySATModel, model)

        #Do some basic calculations to speedup the rest
        self._features = self.model.features
        self._common_features = sat_operations.PySATCoreFeatures().execute(self.model).get_result()
        self._dead_features = sat_operations.PySATDeadFeatures().execute(self.model).get_result()
        # Get all methods that are marked with the metric_method decorator
        metric_methods = [getattr(self, method_name) for method_name in dir(self)
                          if callable(getattr(self, method_name)) and 
                          hasattr(getattr(self, method_name), '_is_metric_method')]
        if self.filter is not None:
            metric_methods = [method for method in metric_methods 
                              if method.__name__ in self.filter]

        return [method() for method in metric_methods]

    @metric_method
    def satisfiable(self) -> dict[str, Any]:
        """A feature model is satisfiable if it represents at least one configuration."""
        if self.model is None:
            raise FlamaException('Model not initialized.')
        name = "Satisfiable"
        _satisfiable = sat_operations.PySATSatisfiable().execute(self.model).get_result()
        result = self.construct_result(name=name, 
                                       doc=self.satisfiable.__doc__, result=_satisfiable)
        return result

    @metric_method
    def core_features(self) -> dict[str, Any]:
        """Features that are part of all the configurations (aka 'common features')."""
        name = "Core features"
        _core = self._common_features
        result = self.construct_result(name=name,
                                       doc=self.core_features.__doc__,
                                       result=_core,
                                       size=len(_core),
                                       ratio=self.get_ratio(_core, self._features, 2))
        return result

    @metric_method
    def variant_features(self) -> dict[str, Any]:
        """Features that do not appear in all the configurations."""
        name = "Variant features"
        _variant_features = [f for f in self._features.values() 
                             if f not in self._common_features and 
                             f not in self._dead_features]
        result = self.construct_result(name=name,
                                       doc=self.variant_features.__doc__,
                                       result=_variant_features,
                                       size=len(_variant_features),
                                       ratio=self.get_ratio(_variant_features, self._features, 2))
        return result

    @metric_method
    def dead_features(self) -> dict[str, Any]:
        """Features that cannot appear in any configuration."""
        if self.model is None:
            raise FlamaException('Model not initialized.')
        name = "Dead features"
        _dead_features = sat_operations.PySATDeadFeatures().execute(self.model).get_result()
        result = self.construct_result(name=name,
                                       doc=self.dead_features.__doc__,
                                       result=_dead_features,
                                       size=len(_dead_features),
                                       ratio=self.get_ratio(_dead_features, self._features, 2))
        return result

    @metric_method
    def unique_features(self) -> dict[str, Any]:
        """Features that appear in exactly one configuration."""
        name = "Unique features"

        seen_once = set()
        seen_multiple = set()

        # Step 2: Iterate through each configuration
        for config in self.configurations()["result"]:
            for feature in config:
                if feature in seen_once:
                    seen_multiple.add(feature)
                else:
                    seen_once.add(feature)

        # Step 3: Find features that are in seen_once but not in seen_multiple
        _unique_features = list(seen_once - seen_multiple)

        result = self.construct_result(name=name,
                                       doc=self.unique_features.__doc__,
                                       result=_unique_features,
                                       size=len(_unique_features),
                                       ratio=self.get_ratio(_unique_features, self._features, 2))
        return result

    @metric_method
    def false_optional_features(self) -> dict[str, Any]:
        """Features defined as optionals the selection of their parents make the feature itself 
        selected as well."""
        if self.model is None:
            raise FlamaException('Model not initialized.')
        name = "False-optional features"
        _fof = sat_operations.PySATFalseOptionalFeatures().execute(self.model).get_result()
        result = self.construct_result(name=name,
                                       doc=self.false_optional_features.__doc__,
                                       result=_fof,
                                       size=len(_fof),
                                       ratio=self.get_ratio(_fof, self._features, 2))
        return result

    @metric_method
    def configurations(self) -> dict[str, Any]:
        """Number of configurations represented by the feature model. 

           If <= is shown, the number represents an upper estimation bound.
        """
        if self.model is None:
            raise FlamaException('Model not initialized.')
        name = "Configurations"
        _configurations = sat_operations.PySATConfigurations().execute(self.model).get_result()
        result = self.construct_result(name=name,
                                       doc=self.configurations.__doc__,
                                       result=_configurations,
                                       size=len(_configurations))
        return result



from flamapy.core.transformations import ModelToText

from flamapy.metamodels.pysat_metamodel.models import PySATModel


class DimacsWriter(ModelToText):

    @staticmethod
    def get_destination_extension() -> str:
        return 'dimacs'

    def __init__(self, path: str, source_model: PySATModel) -> None:
        self.path = path
        self.source_model = source_model

    def transform(self) -> str:
        dimacs_str = pysat_to_dimacs(self.source_model)
        if self.path is not None:
            with open(self.path, 'w', encoding='utf8') as file:
                file.write(dimacs_str)
        return dimacs_str


def pysat_to_dimacs(model: PySATModel) -> str:
    lines = []
    features_dict = model.features
    clauses_list = model.get_all_clauses().clauses
    lines.append(f'p cnf {len(features_dict)} {len(clauses_list)}')
    for identification, name in features_dict.items():
        lines.append(f'c {identification} {name}')
    for clause in clauses_list:
        lines.append(f'{" ".join((str(c) for c in clause))} 0')
    return '\n'.join(lines)



"""
A Java version of this implementation is available at:
https://github.com/HiConfiT/hiconfit-core/blob/main/ca-cdr-package/src/main/java/at/tugraz/ist/ase/cacdr/algorithms/hs/labeler/QuickXPlainLabeler.java
"""

from dataclasses import dataclass
from typing import List

from .labeler import IHSLabelable, LabelerType, AbstractHSParameters
from ...checker import ConsistencyChecker
from ...quickxplain import QuickXPlain


@dataclass
class QuickXPlainParameters(AbstractHSParameters):
    set_b: List[int]

    def __str__(self) -> str:
        return f"QuickXPlainParameters{{C={self.set_c}, B={self.set_b}}}"


class QuickXPlainLabeler(QuickXPlain, IHSLabelable):
    """
    HSLabeler for QuickXPlain algorithm
    """

    def __init__(self, checker: ConsistencyChecker, parameters: QuickXPlainParameters):
        super().__init__(checker)
        self.initial_parameters = parameters

    def get_type(self) -> LabelerType:
        return LabelerType.CONFLICT

    def get_initial_parameters(self) -> AbstractHSParameters:
        return self.initial_parameters

    def get_label(self, parameters: AbstractHSParameters) -> List[List[int]]:
        """
        Identifies a conflict
        """
        assert isinstance(parameters, QuickXPlainParameters), \
            "parameter must be an instance of QuickXPlainParameters"

        set_cs = self.find_conflict(parameters.set_c, parameters.set_b + parameters.set_d)

        if len(set_cs) != 0:
            # reverse the order of the conflict set
            set_cs.reverse()
            return [set_cs]
        return []

    def identify_new_node_parameters(self, param_parent_node: AbstractHSParameters,
                                     arc_label: int) -> AbstractHSParameters:
        """
        Identifies the new node's parameters on the basis of the parent node's parameters.
        """
        assert isinstance(param_parent_node, QuickXPlainParameters), \
            "parameter must be an instance of QuickXPlainParameters"

        new_c = param_parent_node.set_c.copy()
        new_c.remove(arc_label)
        new_b = param_parent_node.set_b.copy()
        new_d = param_parent_node.set_d.copy()
        new_d.append(-1 * arc_label)

        return QuickXPlainParameters(new_c, new_d, new_b)

    def get_instance(self, checker: ConsistencyChecker) -> 'IHSLabelable':
        return QuickXPlainLabeler(checker, self.initial_parameters)



import logging
from typing import Any, cast

from pysat.solvers import Solver

from flamapy.core.operations import FalseOptionalFeatures
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.metamodels.fm_metamodel.models.feature_model import FeatureModel
from flamapy.core.models import VariabilityModel, VariabilityElement
from flamapy.core.exceptions import FlamaException


LOGGER = logging.getLogger('PySATFalseOptionalFeatures')


class PySATFalseOptionalFeatures(FalseOptionalFeatures):

    def __init__(self) -> None:
        self.result: list[Any] = []
        self.solver = Solver(name='glucose3')

    def execute(self, model: VariabilityModel) -> 'PySATFalseOptionalFeatures':
        sat_model = cast(PySATModel, model)
        self.result = self._get_false_optional_features(sat_model)
        return self

    def get_false_optional_features(self) -> list[VariabilityElement]:
        return self.get_result()

    def get_result(self) -> list[Any]:
        return self.result

    def _get_false_optional_features(self, sat_model: PySATModel) -> list[Any]:
        try:
            feature_model = cast(FeatureModel, sat_model.original_model)
        except FlamaException:
            LOGGER.exception("The transformation didn't attach the source model, " 
                             "which is required for this operation.")

        real_optional_features = [f for f in feature_model.get_features()
                                  if not f.is_root() and not f.is_mandatory()]

        result = []
        for clause in sat_model.get_all_clauses():
            self.solver.add_clause(clause)

        for feature in real_optional_features:
            variable = sat_model.variables.get(feature.name)
            parent_feature = feature.get_parent()
            if parent_feature is not None:
                parent_variable = sat_model.variables.get(parent_feature.name)
                assert variable is not None
                satisfiable = self.solver.solve(assumptions=[parent_variable, -variable])
                if not satisfiable:
                    result.append(feature)
        self.solver.delete()
        return result



from abc import abstractmethod
from typing import cast, List, Tuple

from flamapy.core.models import VariabilityModel
from flamapy.core.operations import Operation
from flamapy.metamodels.configuration_metamodel.models import Configuration

from .diagnosis.checker import ConsistencyChecker
from .diagnosis.hsdag.hsdag import HSDAG
from ..models.pysat_diagnosis_model import DiagnosisModel


def _execute_hsdag(model: DiagnosisModel, hsdag: HSDAG) -> Tuple[str, str]:
    hsdag.construct()
    diagnoses = hsdag.get_diagnoses()
    conflicts = hsdag.get_conflicts()

    if len(diagnoses) == 0:
        diag_mess = 'No diagnosis found'
    elif len(diagnoses) == 1:
        diag_mess = 'Diagnosis: '
        diag_mess += model.get_pretty_diagnoses(diagnoses)
    else:
        diag_mess = 'Diagnoses: '
        diag_mess += model.get_pretty_diagnoses(diagnoses)
    if len(conflicts) == 0:
        cs_mess = 'No conflicts found'
    elif len(conflicts) == 1:
        cs_mess = 'Conflict: '
        cs_mess += model.get_pretty_diagnoses(conflicts)
    else:
        cs_mess = 'Conflicts: '
        cs_mess += model.get_pretty_diagnoses(conflicts)

    return cs_mess, diag_mess


class PySATAbstractIdentifier(Operation):
    """
    An abstract operation for computes conflicts or diagnoses.
    Four optional inputs:
    - configuration - a configuration to be diagnosed
    - test_case - a test case to be used for diagnosis
    - max_depth - specify the maximum depth of the HSDAG to be computed
    """

    def __init__(self) -> None:
        self.result = False
        self.configuration = None
        self.test_case = None
        self.solver_name = 'glucose3'
        self.result_messages: List[str] = []

        self.checker = None
        self.max_depth = 0  # 0 means no limit

    def set_max_depth(self, max_depth: int) -> None:
        self.max_depth = max_depth

    def set_configuration(self, configuration: Configuration) -> None:
        self.configuration = configuration

    def set_test_case(self, test_case: Configuration) -> None:
        self.test_case = test_case

    def get_result(self) -> List[str]:
        return self.result_messages

    def execute(self, model: VariabilityModel) -> 'PySATAbstractIdentifier':
        model = cast(DiagnosisModel, model)

        checker, labeler = self.prepare_hsdag(model)

        cs_mess, diag_mess = _execute_hsdag(model, labeler)

        self.set_result_messages(cs_mess, diag_mess)
        checker.delete()
        return self

    @abstractmethod
    def prepare_hsdag(self, model: DiagnosisModel) -> Tuple[ConsistencyChecker, HSDAG]:
        pass

    @abstractmethod
    def set_result_messages(self, cs_mess: str, diag_mess: str) -> None:
        pass


"""
A Java version of this implementation is available at:
https://github.com/HiConfiT/hiconfit-core/blob/main/ca-cdr-package/src/main/java/at/tugraz/ist/ase/cacdr/algorithms/hs/HSDAG.java
"""
from typing import List, Optional, Dict, Union

from .labeler.labeler import IHSLabelable, LabelerType, AbstractHSParameters
from .node import Node, NodeStatus
from .. import utils
from ..utils import diff, contains


class HSDAGException(Exception):
    """
    Exception class for HSDAG
    """


class HSDAG:
    """
    Implementation of the HS-dag algorithm.
    IHSLabeler algorithms could return labels (conflict or diagnosis) which are not minimal.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, labeler: IHSLabelable) -> None:
        self.labeler = labeler  # could be FastDiag or QuickXPlain

        self.max_number_diagnoses = -1  # maximum number of diagnoses to be found
        self.max_number_conflicts = -1  # maximum number of conflicts to be found
        self.max_depth = 0  # maximum depth of the HS-dag

        self.node_labels: List[List[int]] = []  # list of diagnoses or conflicts found
        self.path_labels: List[List[int]] = []  # list of diagnoses or conflicts found

        self.root: Optional[Node] = None  # root node of the HS-dag
        self.open_nodes: List[Node] = []  # list of open nodes
        # Map of <label, list of nodes which have the label as its label>
        self.label_nodes_map: Dict[int, List[Node]] = {}
        self.nodes_lookup: Dict[int, Node] = {}  # Map of <pathLabel, Node>

    def get_conflicts(self) -> List[List[int]]:
        """
        Returns the list of conflicts found.
        """
        if self.labeler.get_type() == LabelerType.CONFLICT:
            return self.node_labels
        return self.path_labels

    def get_diagnoses(self) -> List[List[int]]:
        """
        Returns the list of diagnoses found.
        """
        if self.labeler.get_type() == LabelerType.CONFLICT:
            return self.path_labels
        return self.node_labels

    def should_stop_construction(self) -> bool:
        condition1 = (self.max_number_diagnoses != -1
                      and self.max_number_diagnoses <= len(self.get_diagnoses()))
        condition2 = (self.max_number_conflicts != -1
                      and self.max_number_conflicts <= len(self.get_conflicts()))
        return condition1 or condition2

    def construct(self) -> None:
        """
        Constructs the HS-dag.
        """
        param = self.labeler.get_initial_parameters()

        has_root_label = self.create_root(param)

        if not self.should_stop_construction() and has_root_label:
            self.create_nodes()

    def create_root(self, param: AbstractHSParameters) -> bool:
        """
        Creates the root node of the HS-dag.
        """
        has_root_label = True

        if not self.has_root():
            labels = self.compute_label(self.labeler, param)

            if len(labels) == 0:
                has_root_label = False
            else:  # create root node
                label = self.select_label(labels)
                self.root = Node.create_root(label, param)

                self.open_nodes.append(self.root)

                self.add_node_labels(labels)  # to reuse labels
                self.add_item_to_label_nodes_map(label, self.root)

        return has_root_label

    def create_nodes(self) -> None:
        """
        Creates nodes of the HS-dag.
        """
        while self.has_nodes_to_expand():
            node = self.get_next_node()

            if not node.is_root():
                if self.skip_node(node):
                    continue

                self.label(node)

                if self.should_stop_construction():
                    break

            if node.status == NodeStatus.OPEN:
                self.expand(node)

    def label(self, node: Node) -> None:
        """
        Labels a node - identify new conflict or diagnosis.
        """
        # Reusing labels - H(node) ∩ S = {}, then label node by S
        labels = self.get_reusable_labels(node)

        # compute labels if there are none to reuse
        if len(labels) == 0:
            labels = self.compute_label_from_node(self.labeler, node)

            self.process_labels(labels)

        if len(labels) > 0:
            label = self.select_label(labels)

            node.label = label
            self.add_item_to_label_nodes_map(label, node)

        else:  # found a path label
            self.found_a_path_label_at_node(node)

    def expand(self, node_to_expand: Node) -> None:
        """
        Creates children of a node.
        """
        for arc_label in node_to_expand.label:
            param_parent_node = node_to_expand.parameters
            if param_parent_node is None:
                raise HSDAGException("The parent node must have parameters.")

            new_param = self.labeler.identify_new_node_parameters(param_parent_node, arc_label)

            # rule 1.a - reuse node
            node = self.get_reusable_node(node_to_expand.path_label, arc_label)
            if node is not None:
                node.add_parent(node_to_expand)
            else:  # rule 1.b - generate a new node
                node = Node(parent=node_to_expand, arc_label=arc_label, parameters=new_param)
                hashcode = sum(node.path_label)
                self.nodes_lookup[hashcode] = node

                if not self.can_prune(node):
                    self.open_nodes.append(node)

    def add_item_to_label_nodes_map(self, label: List[int], node: Node) -> None:
        """
        Adds a node to the label_nodes_map.
        """
        hashcode = sum(label)
        if hashcode in self.label_nodes_map:
            self.label_nodes_map[hashcode].append(node)
        else:
            self.label_nodes_map[hashcode] = [node]

    @staticmethod
    def compute_label(labeler: IHSLabelable, param: AbstractHSParameters) -> List[List[int]]:
        return labeler.get_label(param)

    @staticmethod
    def compute_label_from_node(labeler: IHSLabelable, node: Node) -> List[List[int]]:
        param = node.parameters
        if param is None:
            raise HSDAGException("The node must have parameters.")
        return HSDAG.compute_label(labeler, param)

    def add_node_labels(self, labels: List[List[int]]) -> None:
        for label in labels:
            self.node_labels.append(label)

    def found_a_path_label_at_node(self, node: Node) -> None:
        node.status = NodeStatus.CHECKED
        path_label = node.path_label.copy()

        self.path_labels.append(path_label)

    @staticmethod
    def select_label(labels: List[List[int]]) -> List[int]:
        return labels[0]

    def has_nodes_to_expand(self) -> bool:
        return len(self.open_nodes) > 0

    def get_next_node(self) -> Node:
        return self.open_nodes.pop(0)

    def has_root(self) -> bool:
        return self.root is not None

    # Pruning engine
    def skip_node(self, node: Node) -> bool:
        condition1 = self.max_depth != 0 and self.max_depth < node.level
        return node.status != NodeStatus.OPEN or condition1 or self.can_prune(node)

    def can_prune(self, node_2prime: Node) -> bool:
        # 3.i - if n is checked, and n'' is such that H(n) ⊆ H(n'), then close the node n''
        # n is a diagnosis
        for path_label in self.path_labels:
            # if node.path_label.containsAll(path_label):
            if all(elem in path_label for elem in node_2prime.path_label):
                node_2prime.status = NodeStatus.CLOSED
                return True

        # 3.ii - if n has been generated and node n'' is such that H(n') = H(n),
        # then close node n''
        for node in self.open_nodes:
            if len(node.path_label) == len(node_2prime.path_label) \
                    and len(diff(node.path_label, node_2prime.path_label)) == 0:
                node_2prime.status = NodeStatus.CLOSED
                return True
        return False

    def get_reusable_labels(self, node: Node) -> List[List[int]]:
        labels = []
        for label in self.node_labels:
            # H(node) ∩ S = {}
            if not utils.has_intersection(node.path_label, label):
                labels.append(label)
        return labels

    def get_reusable_node(self, path_labels: List[int], arc_label: int) -> Union[Node, None]:
        if len(path_labels) == 0:
            new_path_labels = [arc_label]
        else:
            new_path_labels = path_labels.copy()
            new_path_labels.append(arc_label)
        hashcode = sum(new_path_labels)
        return self.nodes_lookup.get(hashcode)

    def process_labels(self, labels: List[List[int]]) -> None:
        # check existing and obtained labels for subset-relations
        if len(labels) <= 0:
            return

        non_min_labels: List[List[int]] = []
        for first_label in self.node_labels:
            if contains(non_min_labels, first_label):
                continue

            for second_label in labels:
                if contains(non_min_labels, second_label):
                    continue

                greater = first_label if len(first_label) > len(second_label) else second_label
                smaller = second_label if len(first_label) > len(second_label) else first_label
                if not utils.contains_all(greater, smaller):
                    continue

                non_min_labels.append(greater)
                # update the DAG
                nodes = self.label_nodes_map.get(sum(greater))

                if nodes is None:
                    continue

                # get a list of nodes which have the status OPEN
                open_nodes = [n for n in nodes if n.status == NodeStatus.OPEN]

                self.update_dag(greater, open_nodes, smaller)

        # remove the known non - minimal conflicts
        for non_min_label in non_min_labels:
            labels.remove(non_min_label)  # labels.removeAll(non_min_labels)
            hashcode = sum(non_min_label)
            del self.label_nodes_map[hashcode]  # non_min_labels.forEach(label_nodesMap::remove)

        # add new labels to the list of labels
        self.add_node_labels(labels)

    def update_dag(self, greater: List[int], open_nodes: List[Node], smaller: List[int]) -> None:
        for node in open_nodes:
            node.label = smaller  # relabel the node with smaller
            # add new label to the map
            self.add_item_to_label_nodes_map(smaller, node)
            delete = diff(greater, smaller)

            for label in delete:
                child = node.children.get(label)
                if child is None:
                    continue
                child.parents.remove(node)
                del node.children[label]
                self.clean_up_nodes(child)

    def clean_up_nodes(self, node: Node) -> None:
        hashcode = sum(node.path_label)
        del self.nodes_lookup[hashcode]

        if node.status == NodeStatus.OPEN:
            node.status = NodeStatus.PRUNED

        # downward clean up
        for arc_label in node.children.keys():
            child = node.children.get(arc_label)
            if child is not None:
                self.clean_up_nodes(child)



from pysat.formula import CNF

from flamapy.core.exceptions import FlamaException
from flamapy.core.models import VariabilityModel


class PySATModel(VariabilityModel):
    @staticmethod
    def get_extension() -> str:
        return 'pysat'

    def __init__(self) -> None:
        self._cnf = CNF()
        self.variables: dict[str, int] = {}  # feature's name -> id
        self.features: dict[int, str] = {}  # id -> feature's name
        self.original_model: VariabilityModel

    def add_clause(self, clause: list[int]) -> None:
        self._cnf.append(clause)

    def get_variable(self, key: str) -> int:
        if key not in self.variables:
            raise FlamaException(f'Feature {key} not found')
        return self.variables[key]

    def get_all_clauses(self) -> CNF:
        return self._cnf



"""
A Java version of this implementation is available at:
https://github.com/HiConfiT/hiconfit-core/tree/main/ca-cdr-package/src/main/java/at/tugraz/ist/ase/cacdr/algorithms/hs
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import List

from ...checker import ConsistencyChecker


class LabelerType(Enum):
    CONFLICT = 1
    DIAGNOSIS = 2


@dataclass
class AbstractHSParameters:
    set_c: List[int]
    set_d: List[int]


class IHSLabelable(ABC):
    """
    Interface for the HSDAG's labeler
    """

    @abstractmethod
    def get_type(self) -> LabelerType:
        pass

    @abstractmethod
    def get_initial_parameters(self) -> AbstractHSParameters:
        pass

    @abstractmethod
    def get_label(self, parameters: AbstractHSParameters) -> List[List[int]]:
        pass

    @abstractmethod
    def identify_new_node_parameters(self, param_parent_node: AbstractHSParameters,
                                     arc_label: int) -> AbstractHSParameters:
        pass

    @abstractmethod
    def get_instance(self, checker: ConsistencyChecker) -> 'IHSLabelable':
        pass



from typing import cast

from pysat.solvers import Solver

from flamapy.core.operations import Satisfiable

from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.core.models import VariabilityModel


class PySATSatisfiable(Satisfiable):

    def __init__(self) -> None:
        self.result = False
        self.solver = Solver(name='glucose3')

    def is_satisfiable(self) -> bool:
        return self.get_result()

    def get_result(self) -> bool:
        return self.result

    def execute(self, model: VariabilityModel) -> 'PySATSatisfiable':
        sat_model = cast(PySATModel, model)
        self.result = valid(self.solver, sat_model)
        return self


def valid(solver: Solver, model: PySATModel) -> bool:
    for clause in model.get_all_clauses():
        solver.add_clause(clause)
    result = solver.solve()
    solver.delete()
    return result



from typing import List, Dict

from flamapy.metamodels.configuration_metamodel.models import Configuration

from flamapy.metamodels.pysat_metamodel.models import PySATModel
from flamapy.metamodels.fm_metamodel.models.feature_model import Feature


class DiagnosisModel(PySATModel):
    """
    This is a new version of the PySATModel class to support the following tasks:
    1. Diagnosis Task
        If a configuration is given:
            C = configuration
            B = {f0 = true} + CF (i.e., = PySATModel)
        else (no configuration is given):
            a. Diagnosis the feature model
                C = CF (i.e., = PySATModel - {f0 = true})
                B = {f0 = true}
            b. Diagnosis the error
                C = CF (i.e., = PySATModel - {f0 = true})
                B = {f0 = true} + test_case
                where test_case is the following:
                + Dead feature: test_case = {fi = true}
                + False optional feature: test_case = {f_parent = true} & {f_child = false}
    2. Redundancy Detection Task (need negative constraints)
        C = CF (i.e., = PySATModel - {f0 = true})
        B = {}
    """

    @staticmethod
    def get_extension() -> str:
        return 'pysat_diagnosis'

    def __init__(self) -> None:
        super().__init__()
        # set of constraints which could be faulty
        self.set_c: List[int] = []
        # background knowledge (i.e., the knowledge that is known to be true)
        self.set_b: List[int] = []
        # set of all CNF with added assumptions
        self.set_kb: List[List[int]] = []
        # map clauses to relationships/constraint
        self.constraint_map: Dict[str, List[List[int]]] = {}
        # map id of assumptions to relationships/constraint
        self.constraint_assumption_map: Dict[int, str] = {}

    def add_clause_to_map(self, description: str, clauses: List[List[int]]) -> None:
        self.constraint_map[description] = clauses

    def get_c(self) -> List[int]:
        return self.set_c

    def get_b(self) -> List[int]:
        return self.set_b

    def get_kb(self) -> List[List[int]]:
        return self.set_kb

    def get_pretty_diagnoses(self, assumptions: List[List[int]]) -> str:
        diagnoses = []
        for ass in assumptions:
            diag = []
            for item in ass:
                if self.constraint_assumption_map[item]:
                    diag.append(self.constraint_assumption_map[item])
            diagnoses.append(f"[{', '.join(diag)}]")

        return ','.join(diagnoses)

    def prepare_diagnosis_task(self, configuration: Configuration = None,
                               test_case: Configuration = None) -> None:
        """
        Execute this method after the model is built.
        If a configuration is given:
            C = configuration
            B = {f0 = true} + CF (i.e., = PySATModel)
        else (no configuration is given):
            a. Diagnosis the feature model
                C = CF (i.e., = PySATModel - {f0 = true})
                B = {f0 = true}
            b. Diagnosis the error
                C = CF (i.e., = PySATModel - {f0 = true})
                B = {f0 = true} + test_case
                where test_case is the following:
                + Dead feature: test_case = {fi = true}
                + False optional feature: test_case = {f_parent = true} & {f_child = false}
        """
        if configuration is not None:
            # C = configuration
            # B = {f0 = true} + CF (i.e., = PySATModel)
            self._prepare_assumptions(configuration=configuration)
        else:
            if test_case is None:
                # Diagnosis the feature model
                # C = CF (i.e., = PySATModel - {f0 = true})
                # B = {f0 = true}
                self._prepare_assumptions()
            else:
                # Diagnosis the error
                # C = CF (i.e., = PySATModel - {f0 = true})
                # B = {f0 = true} + test_case
                self._prepare_assumptions(test_case=test_case)

    def prepare_redundancy_detection_task(self) -> None:
        """
        This function prepares the model for WipeOutR algorithm.
        Execute this method after the model is built.
        C = CF (i.e., = PySATModel - {f0 = true})
        B = {}
        """
        # C = CF (i.e., = PySATModel - {f0 = true})
        # self.set_c = self._prepare_assumptions
        # self.set_b = []  # B = {}
        # ToDo: TBD

    def _prepare_assumptions(self, configuration: Configuration = None,
                             test_case: Configuration = None) -> None:
        assumption: List[int] = []

        id_assumption = len(self.variables) + 1
        id_assumption = self._prepare_assumptions_for_kb(assumption, id_assumption)

        start_id_configuration = len(assumption)
        if configuration is not None:
            self.constraint_assumption_map = {}  # reset the map
            id_assumption = self._prepare_assumptions_for_configuration(assumption,
                                                                        configuration,
                                                                        id_assumption)

        start_id_test_case = len(assumption)
        if test_case is not None:
            self._prepare_assumptions_for_configuration(assumption, test_case,
                                                        id_assumption)

        if configuration is not None:
            self.set_b = assumption[:start_id_configuration]
            self.set_c = assumption[start_id_configuration:]
        else:
            if test_case is not None:
                self.set_b = [assumption[0]] + assumption[start_id_test_case:]
                self.set_c = assumption[1:start_id_test_case]
            else:
                self.set_b = [assumption[0]]
                self.set_c = assumption[1:]

    def _prepare_assumptions_for_kb(self, assumption: List[int], id_assumption: int) -> int:
        cstr_map = self.constraint_map
        # loop through all tuples in the constraint map
        for _, key in enumerate(cstr_map):
            # get clauses
            clauses = cstr_map[key]
            # loop through all variables in the constraint
            for _, clause in enumerate(clauses):
                # add the assumption variable to the clause
                # assumption => clause
                # i.e., -assumption v clause
                clause.append(-1 * id_assumption)

            assumption.append(id_assumption)
            self.set_kb.extend(clauses)
            self.constraint_assumption_map[id_assumption] = key

            id_assumption += 1

        return id_assumption

    def _convert_keys_to_features(self, configuration: 'Configuration') -> 'Configuration':
        new_elements = {Feature(key) if isinstance(key, str)
                        else key: value for key, value 
                        in configuration.elements.items()}
        return Configuration(new_elements)

    def _prepare_assumptions_for_configuration(self, assumption: List[int],
                                               configuration: Configuration,
                                               id_assumption: int) -> int:
        configuration = self._convert_keys_to_features(configuration)
        config = [feat.name for feat in configuration.elements]
        for feat in config:
            if feat not in self.variables.keys():
                raise KeyError(f'Feature {feat} is not in the model.')

        for feat, value in configuration.elements.items():
            desc = ''
            clause = []

            if value:
                desc = f'{feat.name} = true'
                clause = [self.variables[feat.name], -1 * id_assumption]
            else:
                desc = f'{feat.name} = false'
                clause = [-1 * self.variables[feat.name], -1 * id_assumption]

            assumption.append(id_assumption)
            self.set_kb.append(clause)
            self.constraint_assumption_map[id_assumption] = desc

            id_assumption += 1

        return id_assumption


# pylint: disable=cyclic-import

from .pysat_model import PySATModel
from .txtcnf_model import (
    TextCNFModel,
    CNFLogicConnective,
    TextCNFNotation
)


__all__ = [
    'PySATModel',
    'TextCNFModel',
    'CNFLogicConnective',
    'TextCNFNotation'
]



from .fastdiag_labeler import FastDiagParameters, FastDiagLabeler
from .labeler import IHSLabelable, AbstractHSParameters, LabelerType
from .quickxplain_labeler import QuickXPlainParameters, QuickXPlainLabeler

__all__ = [
    'IHSLabelable',
    'AbstractHSParameters',
    'LabelerType',
    'FastDiagParameters',
    'FastDiagLabeler',
    'QuickXPlainParameters',
    'QuickXPlainLabeler'
]


from typing import Tuple

from flamapy.metamodels.pysat_diagnosis_metamodel.models import DiagnosisModel
from . import PySATAbstractIdentifier
from .diagnosis.checker import ConsistencyChecker
from .diagnosis.hsdag.hsdag import HSDAG
from .diagnosis.hsdag.labeler.quickxplain_labeler import QuickXPlainParameters, QuickXPlainLabeler


class PySATConflict(PySATAbstractIdentifier):
    """
    An operation that computes conflicts and diagnoses
    using the combination of HSDAG and QuickXPlain algorithms.
    Four optional inputs:
    - configuration - a configuration to be diagnosed
    - test_case - a test case to be used for diagnosis
    - max_conflicts - specify the maximum number of conflicts to be computed
    - max_depth - specify the maximum depth of the HSDAG to be computed
    """

    def __init__(self) -> None:
        super().__init__()
        self.max_conflicts = -1  # -1 means no limit

    def set_max_conflicts(self, max_conflicts: int) -> None:
        self.max_conflicts = max_conflicts

    def prepare_hsdag(self, model: DiagnosisModel) -> Tuple[ConsistencyChecker, HSDAG]:
        # transform model to diagnosis model
        model.prepare_diagnosis_task(configuration=self.configuration, test_case=self.test_case)
        # print(f'C: {diag_model.get_C()}')
        # print(f'B: {diag_model.get_B()}')

        set_c = model.get_c()

        # if self.configuration is None:
        # reverse the list to get the correct order of constraints in the diagnosis messages
        #     C.reverse()

        checker = ConsistencyChecker(self.solver_name, model.get_kb())
        parameters = QuickXPlainParameters(set_c, [], model.get_b())
        labeler = QuickXPlainLabeler(checker, parameters)

        hsdag = HSDAG(labeler)
        hsdag.max_number_conflicts = self.max_conflicts
        hsdag.max_depth = self.max_depth

        return checker, hsdag

    def set_result_messages(self, cs_mess: str, diag_mess: str) -> None:
        self.result_messages.append(cs_mess)
        self.result_messages.append(diag_mess)



"""
A Java version of this implementation is available at:
https://github.com/HiConfiT/hiconfit-core/blob/main/ca-cdr-package/src/main/java/at/tugraz/ist/ase/cacdr/algorithms/FastDiagV3.java
"""

import logging
from typing import List

from .checker import ConsistencyChecker
from .utils import split, diff


class FastDiag:
    """
    Implementation of MSS-based FastDiag algorithm.
    Le, V. M., Silva, C. V., Felfernig, A., Benavides, D., Galindo, J., & Tran, T. N. T. (2023).
    FastDiagP: An Algorithm for Parallelized Direct Diagnosis.
    arXiv preprint arXiv:2305.06951.
    """

    def __init__(self, checker: ConsistencyChecker) -> None:
        self.checker = checker

    def find_diagnosis(self, set_c: List[int], set_b: List[int]) -> List[int]:
        """
        Activate FastDiag algorithm if there exists at least one constraint,
        which induces an inconsistency in B. Otherwise, it returns an empty set.

        // Func FastDiag(C, B) : Δ
        // if isEmpty(C) or consistent(B U C) return Φ
        // else return C \\ FD(Φ, C, B)
        :param set_c: a consideration set of constraints
        :param set_b: a background knowledge
        :return: a diagnosis or an empty set
        """
        logging.debug('fastDiag [C=%s, B=%s]', set_c, set_b)
        # print(f'fastDiag [C={C}, B={B}]')

        # if isEmpty(C) or consistent(B U C) return Φ
        if len(set_c) == 0 or self.checker.is_consistent(set_b + set_c, []):
            logging.debug('return Φ')
            # print('return Φ')
            return []

        # return C \ FD(C, B, Φ)
        mss = self._fd([], set_c, set_b)
        diag = diff(set_c, mss)

        logging.debug('return %s', diag)
        # print(f'return {diag}')
        return diag

    def _fd(self, delta: List[int], set_c: List[int], set_b: List[int]) -> List[int]:
        """
        The implementation of MSS-based FastDiag algorithm.
        The algorithm determines a maximal satisfiable subset MSS (Γ) of C U B.

        // Func FD(Δ, C = {c1..cn}, B) : MSS
        // if Δ != Φ and consistent(B U C) return C;
        // if singleton(C) return Φ;
        // k = n/2;
        // C1 = {c1..ck}; C2 = {ck+1..cn};
        // Δ1 = FD(C2, C1, B);
        // Δ2 = FD(C1 - Δ1, C2, B U Δ1);
        // return Δ1 ∪ Δ2;
        :param delta: check to skip redundant consistency checks
        :param set_c: a consideration set of constraints
        :param set_b: a background knowledge
        :return: a maximal satisfiable subset MSS of C U B
        """
        logging.debug('>>> FD [Δ=%s, C=%s, B=%s]', delta, set_c, set_b)

        # if Δ != Φ and consistent(B U C) return C;
        if len(delta) != 0 and self.checker.is_consistent(set_b + set_c, delta):
            logging.debug('<<< return %s', set_c)
            return set_c

        # if singleton(C) return Φ;
        if len(set_c) == 1:
            logging.debug('<<< return Φ')
            return []

        # C1 = {c1..ck}; C2 = {ck+1..cn};
        set_c1, set_c2 = split(set_c)

        # Δ1 = FD(C2, C1, B);
        delta1 = self._fd(set_c2, set_c1, set_b)
        # Δ2 = FD(C1 - Δ1, C2, B U Δ1);
        c1_without_delta1 = diff(set_c1, delta1)
        delta2 = self._fd(c1_without_delta1, set_c2, set_b + delta1)

        logging.debug('<<< return [Δ1={Δ1} ∪ Δ2={Δ2}]')

        # return Δ1 + Δ2
        return delta1 + delta2


from typing import Any, cast

from pysat.solvers import Solver

from flamapy.core.operations import Configurations
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.core.models import VariabilityModel


class PySATConfigurations(Configurations):

    def __init__(self) -> None:
        self.result: list[Configuration] = []
        self.solver = Solver(name='glucose3')

    def get_configurations(self) -> list[Configuration]:
        return self.get_result()

    def get_result(self) -> list[Configuration]:
        return self.result

    def execute(self, model: VariabilityModel) -> 'PySATConfigurations':
        sat_model = cast(PySATModel, model)
        self.result = configurations(self.solver, sat_model)
        return self


def configurations(solver: Solver, model: PySATModel) -> list[Configuration]:
    for clause in model.get_all_clauses():
        solver.add_clause(clause)

    result = []
    for solutions in solver.enum_models():
        product: dict[Any, bool] = {}
        for variable in solutions:
            if variable > 0:
                product[model.features.get(variable)] = True
        result.append(Configuration(product))
    solver.delete()
    return result



# pylint: disable=cyclic-import

from .pysat_abstract_identifier import PySATAbstractIdentifier
from .pysat_conflict import PySATConflict
from .pysat_diagnosis import PySATDiagnosis


__all__ = [
    'PySATAbstractIdentifier',
    'PySATDiagnosis',
    'PySATConflict'
]



from typing import Tuple

from flamapy.metamodels.pysat_diagnosis_metamodel.models import DiagnosisModel
from . import PySATAbstractIdentifier
from .diagnosis.checker import ConsistencyChecker
from .diagnosis.hsdag.hsdag import HSDAG
from .diagnosis.hsdag.labeler.fastdiag_labeler import FastDiagParameters, FastDiagLabeler


class PySATDiagnosis(PySATAbstractIdentifier):
    """
    An operation that computes diagnoses and conflict sets
    using the combination of HSDAG and FastDiag algorithms.
    Four optional inputs:
    - configuration - a configuration to be diagnosed
    - test_case - a test case to be used for diagnosis
    - max_diagnoses - specify the maximum number of diagnoses to be computed
    - max_depth - specify the maximum depth of the HSDAG to be computed
    """

    def __init__(self) -> None:
        super().__init__()
        self.max_diagnoses = -1  # -1 means no limit

    def set_max_diagnoses(self, max_diagnoses: int) -> None:
        self.max_diagnoses = max_diagnoses

    def prepare_hsdag(self, model: DiagnosisModel) -> Tuple[ConsistencyChecker, HSDAG]:
        # transform model to diagnosis model
        model.prepare_diagnosis_task(configuration=self.configuration, test_case=self.test_case)

        set_c = model.get_c()

        checker = ConsistencyChecker(self.solver_name, model.get_kb())
        parameters = FastDiagParameters(set_c, [], model.get_b())
        labeler = FastDiagLabeler(checker, parameters)

        hsdag = HSDAG(labeler)
        hsdag.max_number_diagnoses = self.max_diagnoses
        hsdag.max_depth = self.max_depth

        return checker, hsdag

    def set_result_messages(self, cs_mess: str, diag_mess: str) -> None:
        self.result_messages.append(diag_mess)
        self.result_messages.append(cs_mess)



"""
A Java version of this implementation is available at:
https://github.com/HiConfiT/hiconfit-core/blob/main/ca-cdr-package/src/main/java/at/tugraz/ist/ase/cacdr/algorithms/hs/labeler/FastDiagV3Labeler.java
"""

from dataclasses import dataclass
from typing import List

from .labeler import IHSLabelable, LabelerType, AbstractHSParameters
from ...checker import ConsistencyChecker
from ...fastdiag import FastDiag


@dataclass
class FastDiagParameters(AbstractHSParameters):
    set_b: List[int]

    def __str__(self) -> str:
        return f"FastDiagParameters{{C={self.set_c}, B={self.set_b}}}"


class FastDiagLabeler(FastDiag, IHSLabelable):
    """
    HSLabeler for FastDiag algorithm
    """

    def __init__(self, checker: ConsistencyChecker, parameters: FastDiagParameters):
        super().__init__(checker)
        self.initial_parameters = parameters

    def get_type(self) -> LabelerType:
        return LabelerType.DIAGNOSIS

    def get_initial_parameters(self) -> AbstractHSParameters:
        return self.initial_parameters

    def get_label(self, parameters: AbstractHSParameters) -> List[List[int]]:
        """
        Identifies a diagnosis
        """
        assert isinstance(parameters, FastDiagParameters), \
            "parameter must be an instance of FastDiagParameters"
        neg_c = [-1 * item for item in parameters.set_c]
        if len(parameters.set_c) >= 1 \
                and (len(parameters.set_b) == 0
                     or self.checker.is_consistent(parameters.set_b + neg_c, [])):

            diag = self.find_diagnosis(parameters.set_c, parameters.set_b)
            if len(diag) != 0:
                return [diag]
        return []

    def identify_new_node_parameters(self, param_parent_node: AbstractHSParameters,
                                     arc_label: int) -> AbstractHSParameters:
        """
        Identifies the new node's parameters on the basis of the parent node's parameters.
        """
        assert isinstance(param_parent_node, FastDiagParameters),\
            "parameter must be an instance of FastDiagParameters"

        new_c = param_parent_node.set_c.copy()
        new_c.remove(arc_label)
        new_b = param_parent_node.set_b.copy()
        new_b.append(arc_label)
        # D = param_parent_node.D.copy()
        # D.append(arcLabel)

        return FastDiagParameters(new_c, [], new_b)

    def get_instance(self, checker: ConsistencyChecker) -> IHSLabelable:
        return FastDiagLabeler(checker, self.initial_parameters)



"""
A Java version of this implementation is available at:
https://github.com/HiConfiT/hiconfit-core/blob/main/ca-cdr-package/src/main/java/at/tugraz/ist/ase/cacdr/algorithms/hs/Node.java
"""
from enum import Enum
from typing import Optional, List, Any, Dict

from .labeler.labeler import AbstractHSParameters


class NodeStatus(Enum):
    OPEN = 'Open'
    CLOSED = 'Closed'
    PRUNED = 'Pruned'
    CHECKED = 'Checked'  # Checked - the label of this node is a Conflict or a Diagnosis


# pylint: disable=too-many-instance-attributes
class Node:
    """
    A data structure representing a node of an HS-dag.
    """
    generating_node_id = -1

    def __init__(self, parent: Optional['Node'] = None, arc_label: Optional[int] = None,
                 parameters: Optional[AbstractHSParameters] = None):
        self.node_id = Node.generating_node_id = Node.generating_node_id + 1
        self.status = NodeStatus.OPEN  # node status
        # label of the node - it can be a minimal conflict or a diagnosis or null
        self.label: List[Any] = []
        # the constraint associated to the arch which comes to this node
        self.arc_label = arc_label
        self.parameters = parameters
        self.path_label: List[int] = []  # labels of the path to here
        self.parents: List[Node] = []  # the node's parents. Can be null for the root node.
        self.children: Dict[Any, Any] = {}  # the node's children

        if parent is None:
            self.level = 0  # tree level
        else:
            self.level = parent.level + 1  # tree level
            self.parents = []
            self.parents.append(parent)
            if len(parent.path_label) == 0:
                self.path_label = []
            else:
                self.path_label = parent.path_label.copy()
            if arc_label is not None:
                self.path_label.append(arc_label)

            parent.children[arc_label] = self

    @staticmethod
    def create_root(label: List[int], parameters: AbstractHSParameters) -> 'Node':
        Node.generating_node_id = -1

        root = Node()
        root.label = label
        root.parameters = parameters

        return root

    def add_parent(self, parent: 'Node') -> None:
        if self.is_root():
            raise ValueError("The root node cannot have parents.")
        self.parents.append(parent)

    def add_child(self, arc_label: int, child: 'Node') -> None:
        self.children[arc_label] = child
        child.add_parent(self)

    def is_root(self) -> bool:
        return len(self.parents) == 0

    def __str__(self) -> str:
        return (f"Node{{id={self.node_id}, level={self.level}, status={self.status}, "
                f"label={self.label}, parameter={self.parameters}, arcLabel={self.arc_label}, "
                f"pathLabels={self.path_label}}}")



from flamapy.core.operations import Commonality
from flamapy.core.models import VariabilityModel
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from .pysat_configurations import PySATConfigurations


class PySATCommonality(Commonality):

    def __init__(self) -> None:
        self.result: float = 0
        self.configuration = Configuration({})

    def set_configuration(self, configuration: Configuration) -> None:
        self.configuration = configuration

    def get_commonality(self) -> float:
        return self.get_result()

    def get_result(self) -> float:
        return self.result

    def execute(self, model: VariabilityModel) -> 'PySATCommonality':
        pysat_products_op = PySATConfigurations()
        pysat_products_op.execute(model)
        products = pysat_products_op.get_result()

        feature = list(self.configuration.elements.keys())[0]

        count = 0
        for product in products:
            count = count + 1 if feature in product else count

        self.result = count / len(products)

        return self



import itertools
from typing import Any, List

from flamapy.core.transformations import ModelToModel
from flamapy.metamodels.fm_metamodel.models.feature_model import (
    FeatureModel,
    Constraint,
    Feature,
    Relation,
)
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class FmToPysat(ModelToModel):
    @staticmethod
    def get_source_extension() -> str:
        return 'fm'

    @staticmethod
    def get_destination_extension() -> str:
        return 'pysat'

    def __init__(self, source_model: FeatureModel) -> None:
        self.source_model = source_model
        self.counter = 1
        self.destination_model = PySATModel()
        self.destination_model.original_model = source_model
        # self.r_cnf = self.destination_model.r_cnf
        # self.ctc_cnf = self.destination_model.ctc_cnf

    def add_feature(self, feature: Feature) -> None:
        if feature.name not in self.destination_model.variables:
            self.destination_model.variables[feature.name] = self.counter
            self.destination_model.features[self.counter] = feature.name
            self.counter += 1

    def add_root(self, feature: Feature) -> None:
        # self.r_cnf.append([self.destination_model.variables.get(feature.name)])
        value = self.destination_model.get_variable(feature.name)
        self.destination_model.add_clause([value])

    def _add_mandatory_relation(self, relation: Relation) -> List[List[int]]:
        value_parent = self.destination_model.get_variable(relation.parent.name)
        value_child = self.destination_model.get_variable(relation.children[0].name)
        clauses = [[-1 * value_parent, value_child], [-1 * value_child, value_parent]]
        return clauses

    def _add_optional_relation(self, relation: Relation) -> List[List[int]]:
        value_parent = self.destination_model.get_variable(relation.parent.name)
        value_children = self.destination_model.get_variable(relation.children[0].name)
        clauses = [[-1 * value_children, value_parent]]
        return clauses

    def _add_or_relation(self, relation: Relation) -> List[List[int]]:
        # this is a 1 to n relatinship with multiple childs
        # add the first cnf child1 or child2 or ... or childN or no parent)
        # first elem of the constraint
        value_parent = self.destination_model.get_variable(relation.parent.name)

        alt_cnf = [-1 * value_parent]
        for child in relation.children:
            alt_cnf.append(self.destination_model.get_variable(child.name))
        clauses = [alt_cnf]

        for child in relation.children:
            clauses.append([
                -1 * self.destination_model.get_variable(child.name),
                value_parent
            ])

        return clauses

    def _add_alternative_relation(self, relation: Relation) -> List[List[int]]:
        # this is a 1 to 1 relatinship with multiple childs
        # add the first cnf child1 or child2 or ... or childN or no parent)

        value_parent = self.destination_model.get_variable(relation.parent.name)
        # first elem of the constraint
        alt_cnf = [-1 * value_parent]
        for child in relation.children:
            alt_cnf.append(self.destination_model.get_variable(child.name))
        clauses = [alt_cnf]

        for i, _ in enumerate(relation.children):
            for j in range(i + 1, len(relation.children)):
                if i != j:
                    clauses.append([
                        -1 * self.destination_model.get_variable(relation.children[i].name),
                        -1 * self.destination_model.get_variable(relation.children[j].name)
                    ])
            clauses.append([
                -1 * self.destination_model.get_variable(relation.children[i].name),
                value_parent
            ])
        return clauses

    def _add_constraint_relation(self, relation: Relation) -> List[List[int]]:
        value_parent = self.destination_model.get_variable(relation.parent.name)

        # This is a _min to _max relationship
        _min = relation.card_min
        _max = relation.card_max

        clauses = []

        for val in range(len(relation.children) + 1):
            if val < _min or val > _max:
                # combinations of val elements
                for combination in itertools.combinations(relation.children, val):
                    cnf = [-1 * value_parent]
                    for feat in relation.children:
                        if feat in combination:
                            cnf.append(-1 * self.destination_model.get_variable(feat.name))
                        else:
                            cnf.append(self.destination_model.get_variable(feat.name))
                    clauses.append(cnf)

        # there is a special case when coping with the upper part of the thru table
        # In the case of allowing 0 childs, you cannot exclude the option  in that
        # no feature in this relation is activated
        for val in range(1, len(relation.children) + 1):

            for combination in itertools.combinations(relation.children, val):
                cnf = [value_parent]
                for feat in relation.children:
                    if feat in combination:
                        cnf.append(-1 * self.destination_model.get_variable(feat.name))
                    else:
                        cnf.append(self.destination_model.get_variable(feat.name))
                clauses.append(cnf)
        return clauses

    def _store_constraint_clauses(self, clauses: List[List[int]]) -> None:
        for clause in clauses:
            self.destination_model.add_clause(clause)

    def add_relation(self, relation: Relation) -> None:
        if relation.is_mandatory():
            clauses = self._add_mandatory_relation(relation)
        elif relation.is_optional():
            clauses = self._add_optional_relation(relation)
        elif relation.is_or():  
            clauses = self._add_or_relation(relation)
        elif relation.is_alternative():  
            clauses = self._add_alternative_relation(relation)
        else:
            clauses = self._add_constraint_relation(relation)
        self._store_constraint_clauses(clauses)

    def add_constraint(self, ctc: Constraint) -> None:
        def get_term_variable(term: Any) -> int:
            if term.startswith('-'):
                return -self.destination_model.get_variable(term[1:])

            return self.destination_model.get_variable(term)

        clauses = ctc.ast.get_clauses()
        for clause in clauses:
            clause_variables = list(map(get_term_variable, clause))
            self.destination_model.add_clause(clause_variables)

    def transform(self) -> PySATModel:
        for feature in self.source_model.get_features():
            self.add_feature(feature)

        self.add_root(self.source_model.root)

        for relation in self.source_model.get_relations():
            self.add_relation(relation)

        for constraint in self.source_model.get_constraints():
            self.add_constraint(constraint)

        return self.destination_model



from .pysat_satisfiable import PySATSatisfiable
from .pysat_satisfiable_configuration import PySATSatisfiableConfiguration
from .pysat_configurations import PySATConfigurations
from .pysat_configurations_number import PySATConfigurationsNumber
from .pysat_commonality import PySATCommonality
from .pysat_filter import PySATFilter
from .pysat_core_features import PySATCoreFeatures
from .pysat_dead_features import PySATDeadFeatures
from .pysat_false_optional_features import PySATFalseOptionalFeatures
from .pysat_metrics import PySATMetrics


__all__ = [
    'PySATSatisfiable',
    'PySATSatisfiableConfiguration',
    'PySATConfigurations',
    'PySATConfigurationsNumber',
    'PySATCommonality',
    'PySATFilter',
    'PySATCoreFeatures',
    'PySATDeadFeatures',
    'PySATFalseOptionalFeatures',
    'PySATMetrics',
]



"""
A Java version of this implementation is available at:
https://github.com/HiConfiT/hiconfit-core/blob/main/ca-cdr-package/src/main/java/at/tugraz/ist/ase/cacdr/checker/ChocoConsistencyChecker.java
"""
from typing import List

from pysat.solvers import Solver


class ConsistencyChecker:

    def __init__(self, solver_name: str, set_kb: List[List[int]]) -> None:
        self.result = False

        self.solver = Solver(solver_name, bootstrap_with=set_kb)

    def is_consistent(self, set_c: List[int], delta: List[int]) -> bool:
        """
        Check if the given CNF formula is consistent using a solver.
        :param set_c: a list of assumptions should be added to the CNF formula
        :param delta: a list of assumptions should not be added to the CNF formula
        :return: a boolean value indicating whether the given CNF formula is consistent
        """
        assumptions = set_c + [-1 * item for item in delta]
        self.result = self.solver.solve(assumptions=assumptions)
        # print(f"assumptions: {assumptions} - result: {self.result}")
        return self.result

    def delete(self) -> None:
        self.solver.delete()


from typing import cast

from pysat.solvers import Solver

from flamapy.core.models import VariabilityModel
from flamapy.core.operations import SatisfiableConfiguration
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class PySATSatisfiableConfiguration(SatisfiableConfiguration):

    def __init__(self) -> None:
        self.result = False
        self.configuration = Configuration(elements={})
        self.solver = Solver(name='glucose3')

    def is_satisfiable(self) -> bool:
        return self.result

    def get_result(self) -> bool:
        return self.is_satisfiable()

    def set_configuration(self, configuration: Configuration) -> None:
        self.configuration = configuration

    def execute(self, model: VariabilityModel) -> 'PySATSatisfiableConfiguration':
        sat_model = cast(PySATModel, model)

        for clause in sat_model.get_all_clauses():  # AC es conjunto de conjuntos
            self.solver.add_clause(clause)  # añadimos la constraint

        if not self.configuration.is_full:
            assumptions = []
            for feature, selected in self.configuration.elements.items():
                if selected:
                    assumptions.append(sat_model.variables[feature])
                else:
                    assumptions.append(-sat_model.variables[feature])
        else:
            missing_features = [feature for feature in self.configuration.elements.keys() 
                                if feature not in sat_model.variables.keys()]

            if missing_features:
                print("The features that are missing are:", list(missing_features))
                print("The feature model contains the following features:", 
                      list(sat_model.variables.keys()))
                self.result = False
                return self

            assumptions = []
            for feature in sat_model.features.values():

                if self.configuration.elements.get(feature, False):
                    assumptions.append(sat_model.variables[feature])
                else:
                    assumptions.append(-sat_model.variables[feature])

        self.result = self.solver.solve(assumptions=assumptions)
        self.solver.delete()
        return self


from flamapy.core.transformations import TextToModel

from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.metamodels.pysat_metamodel.models.txtcnf_model import (
    TextCNFModel,
    TextCNFNotation,
    CNFLogicConnective
)


class CNFReader(TextToModel):
    """
    Read a CNF formula as a string representing a feature model.

    The expected format is the generated by FeatureIDE when exporting the model as CNF (.txt).
    That generates a file with the CNF formula in four different notations:
        Logical Symbols:
            (A) ∧ (¬B ∨ C) ∧ ...
        Textual Symbols:
            (A) and (not B or C) and ...
        Java Symbols:
            (A) && (!B || C) && ...
        Short Symbols:
            (A) & (-B | C) & ...
    This class is able to read any of these notations, but only one notation,
    so the .txt file should be modified to include only
    one of those notations by removing the others.
    """

    @staticmethod
    def get_source_extension() -> str:
        return 'txt'

    def __init__(self, path: str):
        self._path = path
        self.counter = 1
        self.destination_model = PySATModel()

    def transform(self) -> PySATModel:
        cnf_model = TextCNFModel()
        cnf_model.from_textual_cnf_file(self._path)
        cnf_notation = cnf_model.get_textual_cnf_notation()
        cnf_formula = cnf_model.get_textual_cnf_formula(cnf_notation)

        self._extract_clauses(cnf_formula, cnf_notation)
        return self.destination_model

    def _add_feature(self, feature_name: str) -> None:
        if feature_name not in self.destination_model.variables:
            self.destination_model.variables[feature_name] = self.counter
            self.destination_model.features[self.counter] = feature_name
            self.counter += 1

    def _extract_clauses(self, cnf_formula: str, cnf_notation: TextCNFNotation) -> None:
        and_symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.AND] + ' '
        clauses = list(map(lambda c: c[1:len(c) - 1], cnf_formula.split(and_symbol_pattern)))
        # Remove initial and final parenthesis

        # Remove final parenthesis of last clause (because of the possible end of line: '\n')
        if ')' in clauses[len(clauses) - 1]:
            clauses[len(clauses) - 1] = clauses[len(clauses) - 1][:-1]

        for _c in clauses:
            tokens = flamapy.metamodels.pysat_metamodel.operations.diagnosis.utils.split(' ')  # type: ignore[name-defined] # noqa: F821, E501
            tokens = list(filter(lambda t: t != cnf_notation.value[CNFLogicConnective.OR], tokens))
            logic_not = False
            cnf_clause = []
            for feature in tokens:
                if feature == cnf_notation.value[CNFLogicConnective.NOT]:
                    logic_not = True
                elif feature.startswith(cnf_notation.value[CNFLogicConnective.NOT]):
                    feature = feature.replace(cnf_notation.value[CNFLogicConnective.NOT], '', 1)
                    self._add_feature(feature)
                    cnf_clause.append(-1 * self.destination_model.variables[feature])
                else:
                    self._add_feature(feature)
                    if logic_not:
                        cnf_clause.append(-1 * self.destination_model.variables[feature])
                    else:
                        cnf_clause.append(self.destination_model.variables[feature])
                    logic_not = False
            self.destination_model.add_clause(cnf_clause)

    # def get_cnf_formula(self, cnf_output_syntax: TextCNFNotation = TextCNFNotation.JAVA) -> str:
    #     cnf_formula = self._read_cnf_formula()
    #     cnf_notation = identify_notation(cnf_formula)

    #     if cnf_output_syntax == cnf_notation:
    #         return cnf_formula
    #     # Translate AND operators
    #     symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.AND] + ' '
    #     new_symbol = ' ' + cnf_output_syntax.value[CNFLogicConnective.AND] + ' '
    #     cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

    #     # Translate OR operators
    #     symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.OR] + ' '
    #     new_symbol = ' ' + cnf_output_syntax.value[CNFLogicConnective.OR] + ' '
    #     cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

    #     # Translate NOT operators (this is more complex because
    #     # the symbol may be part of a feature's name)
    #     if cnf_notation == TextCNFNotation.TEXTUAL:
    #         symbol_pattern = cnf_notation.value[CNFLogicConnective.NOT] + ' '
    #         new_symbol = cnf_output_syntax.value[CNFLogicConnective.NOT]
    #         cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
    #     elif cnf_output_syntax == TextCNFNotation.TEXTUAL:
    #         symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.NOT]
    #         new_symbol = ' ' + cnf_output_syntax.value[CNFLogicConnective.NOT] + ' '
    #         cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

    #         symbol_pattern = '(' + cnf_notation.value[CNFLogicConnective.NOT]
    #         new_symbol = '(' + cnf_output_syntax.value[CNFLogicConnective.NOT] + ' '
    #         cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
    #     else:
    #         symbol_pattern = ' ' + cnf_notation.value[CNFLogicConnective.NOT]
    #         new_symbol = ' ' + cnf_output_syntax.value[CNFLogicConnective.NOT]
    #         cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)

    #         symbol_pattern = '(' + cnf_notation.value[CNFLogicConnective.NOT]
    #         new_symbol = '(' + cnf_output_syntax.value[CNFLogicConnective.NOT]
    #         cnf_formula = cnf_formula.replace(symbol_pattern, new_symbol)
    #     return cnf_formula



from typing import Any, cast, Optional

from pysat.solvers import Solver

from flamapy.core.models import VariabilityModel
from flamapy.core.operations import Sampling
from flamapy.core.exceptions import FlamaException
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class PySATSampling(Sampling):

    def __init__(self) -> None:
        self.result: list[Configuration] = []
        self.sample_size: int = 0
        self.with_replacement: bool = False
        self.partial_configuration: Configuration = None
        self.solver = Solver(name='glucose3')

    def set_sample_size(self, sample_size: int) -> None:
        if sample_size < 0:
            raise FlamaException(f'Sample size {sample_size} cannot be negative.')
        self.sample_size = sample_size

    def set_with_replacement(self, with_replacement: bool) -> None:
        self.with_replacement = with_replacement

    def set_partial_configuration(self, partial_configuration: Configuration) -> None:
        self.partial_configuration = partial_configuration

    def get_sample(self) -> list[Configuration]:
        return self.get_result()

    def get_result(self) -> list[Configuration]:
        return self.result

    def execute(self, model: VariabilityModel) -> 'PySATSampling':
        sat_model = cast(PySATModel, model)
        self.result = sample(self.solver, 
                             sat_model, 
                             self.sample_size, 
                             self.with_replacement, 
                             self.partial_configuration)
        return self


def sample(solver: Solver,
           model: PySATModel, 
           sample_size: int, 
           with_replacement: bool,  # pylint: disable=unused-argument
           partial_configuration: Optional[Configuration]  # pylint: disable=unused-argument
           ) -> list[Configuration]:
    if sample_size == 0:
        return []

    for clause in model.get_all_clauses():
        solver.add_clause(clause)

    products = []
    for solutions in solver.enum_models():
        product: dict[Any, bool] = {}
        for variable in solutions:
            if variable > 0:
                product[model.features.get(variable)] = True
        products.append(Configuration(product))
        if len(products) == sample_size:
            solver.delete()
            return products
    solver.delete()
    return products



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

def read_requirements(file):
    with open(file, "r") as fh:
        return fh.read().splitlines()

# Read requirements from the requirements.txt file
requirements = read_requirements("requirements.txt")

# Read development requirements from the dev-requirements.txt file
dev_requirements = read_requirements("requirements-dev.txt")

setuptools.setup(
    name="flamapy-sat",
    version="2.0.1",
    author="Flamapy",
    author_email="flamapy@us.es",
    description="flamapy-sat is a plugin to flamapy module",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/flamapy/pysat_metamodel",
    packages=setuptools.find_namespace_packages(include=['flamapy.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=requirements,
    extras_require={
        'dev':dev_requirements
    }
)



#!/usr/bin/env python
"""Provides utility functions."""
from typing import List, Tuple


def split(clauses: List[int]) -> Tuple[List[int], List[int]]:
    """
    Splits the given list of constraints/clauses into two parts.
    :param clauses: a list of clauses
    :return: a tuple of two lists
    """
    half_size = len(clauses) // 2
    return clauses[:half_size], clauses[half_size:]


def diff(list_x: List[int], list_y: List[int]) -> List[int]:
    """
    Returns the difference of two lists.
    :param list_x: list
    :param list_y: list
    :return: list
    """
    return [item for item in list_x if item not in list_y]


def get_hashcode(clauses: List[int]) -> str:
    """
    Returns the hashcode of the given CNF formula.
    :param clauses: a list of clauses
    :return: the hashcode of the given CNF formula
    """
    # clauses = sorted(clauses, key=lambda x: x[0])
    clauses = sorted(clauses)
    return str(clauses)


def has_intersection(list1: List[int], list2: List[int]) -> bool:
    return any(i in list1 for i in list2)


def contains(list_of_lists: List[List[int]], a_list: List[int]) -> bool:
    return any(a_list == x for x in list_of_lists)


def contains_all(greater: List[int], smaller: List[int]) -> bool:
    return all(i in smaller for i in greater)



"""
A Java version of this implementation is available at:
https://github.com/HiConfiT/hiconfit-core/blob/main/ca-cdr-package/src/main/java/at/tugraz/ist/ase/cacdr/algorithms/QuickXPlain.java
"""

import logging
from typing import List

from .checker import ConsistencyChecker
from .utils import split


class QuickXPlain:
    """
    Implementation of QuickXPlain algorithm
    Junker, Ulrich. "Quickxplain: Conflict detection
    for arbitrary constraint propagation algorithms."
    IJCAI’01 Workshop on Modelling and Solving problems with constraints. Vol. 4. 2001.
    """

    def __init__(self, checker: ConsistencyChecker) -> None:
        self.checker = checker

    def find_conflict(self, set_c: List[int], set_b: List[int]) -> List[int]:
        """
        // Func QuickXPlain(C={c1,c2,…, cm}, B): CS
        // IF consistent(B∪C) return "No conflict";
        // IF isEmpty(C) return Φ;
        // ELSE return QX(Φ, C, B);
        :param set_c: a consideration set
        :param set_b: a background knowledge
        :return: a conflict set or an empty set
        """
        logging.debug('>>> QuickXPlain [C=%s, B=%s]', set_c, set_b)
        # print(f'quickXPlain [C={C}, B={B}]')

        # if C is empty or consistent(B U C) then return empty set
        if len(set_c) == 0 or self.checker.is_consistent(set_b + set_c, []):
            logging.debug('return Φ')
            # print('return Φ')
            return []

        # return QX(Φ, C, B)
        set_cs = self._qx([], set_c, set_b)

        logging.debug('return %s', set_cs)
        # print(f'return {cs}')
        return set_cs

    def _qx(self, set_d: List[int], set_c: List[int], set_b: List[int]) -> List[int]:
        """
        // func QX(Δ, C={c1,c2, …, cq}, B): CS
        // IF (Δ != Φ AND inconsistent(B)) return Φ;
        // IF singleton(C) return C;
        // k = q/2;
        // C1 <-- {c1, …, ck}; C2 <-- {ck+1, …, cq};
        // CS1 <-- QX(C2, C1, B ∪ C2);
        // CS2 <-- QX(CS1, C2, B ∪ CS1);
        // return (CS1 ∪ CS2)
        :param set_d: check to skip redundant consistency checks
        :param set_c: a consideration set of constraints
        :param set_b: a background knowledge
        :return: a conflict set or an empty set
        """
        logging.debug('>>> QX [D=%s, C=%s, B=%s]', set_d, set_c, set_b)

        # if D != Φ and inconsistent(B) then return Φ
        if len(set_d) != 0 and not self.checker.is_consistent(set_b, set_c):
            logging.debug('<<< return Φ')
            return []

        # if C is singleton then return C
        if len(set_c) == 1:
            logging.debug('<<< return %s', set_c)
            return set_c

        # C1 = {c1..ck}; C2 = {ck+1..cn};
        set_c1, set_c2 = split(set_c)

        # CS1 = QX(C2, C1, B U C2)
        cs1 = self._qx(set_c2, set_c1, (set_b + set_c2))
        # CS2 = QX(CS1, C2, B U CS1)
        cs2 = self._qx(cs1, set_c2, (set_b + cs1))

        logging.debug('<<< return %s', (cs1 + cs2))

        # return CS1 U CS2
        return cs1 + cs2
